package cz.cuni.amis.pogamut.ut2004.lectures.hsbot;

import cz.cuni.amis.pathfinding.alg.astar.AStarResult;
import cz.cuni.amis.pathfinding.map.IPFMapView;
import cz.cuni.amis.pogamut.base.agent.navigation.impl.PrecomputedPathFuture;
import cz.cuni.amis.pogamut.base.utils.guice.AgentScoped;
import cz.cuni.amis.pogamut.base.utils.math.DistanceUtils;
import cz.cuni.amis.pogamut.base3d.worldview.object.ILocated;
import cz.cuni.amis.pogamut.base3d.worldview.object.Location;
import cz.cuni.amis.pogamut.ut2004.agent.module.sensor.visibility.model.VisibilityLocation;
import cz.cuni.amis.pogamut.ut2004.agent.module.utils.TabooSet;
import cz.cuni.amis.pogamut.ut2004.agent.navigation.NavigationState;
import cz.cuni.amis.pogamut.ut2004.bot.impl.UT2004Bot;
import cz.cuni.amis.pogamut.ut2004.communication.messages.gbcommands.Initialize;
import cz.cuni.amis.pogamut.ut2004.communication.messages.gbinfomessages.*;
import cz.cuni.amis.pogamut.ut2004.hideandseek.bot.UT2004BotHSController;
import cz.cuni.amis.pogamut.ut2004.hideandseek.protocol.HSBotState;
import cz.cuni.amis.pogamut.ut2004.hideandseek.protocol.messages.*;
import cz.cuni.amis.pogamut.ut2004.utils.UT2004BotRunner;
import cz.cuni.amis.utils.IFilter;
import cz.cuni.amis.utils.collections.MyCollections;
import cz.cuni.amis.utils.exception.PogamutException;
import cz.cuni.amis.utils.flag.FlagListener;
import java.util.ArrayList;
import java.util.Collection;
import java.util.logging.Level;

/**
 * HideAndSeek bot torso
 *
 * @author Jimmy
 */
@AgentScoped
public class HideAndSeekBot extends UT2004BotHSController<UT2004Bot> {

    /**
     * Just for the numbering of bots.
     */
    private static int number = 0;

    @Override
    public Initialize getInitializeCommand() {
        return new Initialize().setName("Milan_Rybar" + (++number));
    }

    @SuppressWarnings("unchecked")
    @Override
    public void botInitialized(GameInfo gameInfo, ConfigChange config, InitedMessage init) {
        navigation.getLog().setLevel(Level.WARNING);
    }

    @Override
    public void botFirstSpawn(GameInfo gameInfo, ConfigChange config, InitedMessage init, Self self) {
        // change to ALL to receive logs from the navigation so you can get a grasp on how it is working
        pathExecutor.getLog().setLevel(Level.WARNING);
    }

    // --------------------
    // ====================
    // HIDE AND SEEK EVENTS
    // ====================
    // --------------------
    
    private int runnerSurvived = 0;
    private int runnerSafe = 0;
    private int runnerMeGood = 0;
    private double roundTime = 0;
    
    /**
     * Some RUNNER has just survived the round.
     * <p><p>
     * This event is triggered at the end of the round for every RUNNER that has not been FOULED, CAPTURED and did not make it to SAFE area.
     *
     * @param event
     */
    @Override
    protected void hsRunnerSurvived(HSRunnerSurvived event, boolean me) {
        ++runnerSurvived;
    }

    /**
     * Some RUNNER has just been spotted by the SEEKER.
     * @param event
     */
    @Override
    protected void hsRunnerSpotted(HSRunnerSpotted event, boolean me) {
        // runner - seeker pobezi na pikolu, takze se radeji jeste schovame
        if (hide.isMeRunner() && hide.isMeAlive() && !me) {
            runnerRunToHide(4 * 5);
        }

        // seeker - bezime na pikolu
        if (hide.isMeSeeker() && hide.isMeAlive()) {
            if (playerToCapture == this.players.getPlayer(event.getBotId())) {
                playerToCapture = null;
            }
            seekerRunToSafeArea();
        }
    }

    /**
     * Some RUNNER has just made it to the safe-area.
     * @param event
     */
    @Override
    protected void hsRunnerSafe(HSRunnerSafe event, boolean me) {
        ++runnerSafe;
        
        // runner - je pravdepodobne (mozne), ze seeker pobezi na pikolu, takze se radeji jeste schovame
        if (hide.isMeRunner() && hide.isMeAlive() && !me) {
            runnerRunToHide(4 * 2);
        }
        // runner - zachranil jsem se
        else if (hide.isMeRunner() && hide.isMeAlive() && me) {
            ++runnerMeGood;
        }

        // seeker - koukneme jestli jsme runnera pronasledovali a pripadne pobezime na pikolu
        if (hide.isMeSeeker() && hide.isMeAlive()) {
            if (playerToCapture == this.players.getPlayer(event.getBotId())) {
                playerToCapture = null;
            } else {
                seekerRunToSafeArea();
            }
        }
    }

    /**
     * Some RUNNER has just been fouled-out because it stepped into the restricted-area while it was activated.
     * @param event
     * @param me whether this event speaks about your bot
     */
    @Override
    protected void hsRunnerFouled(HSRunnerFouled event, boolean me) {
        if (hide.isMeSeeker()) {
            if (playerToCapture == this.players.getPlayer(event.getBotId())) {
                playerToCapture = null;
            }
        }
    }

    /**
     * Some RUNNER has just been captured by the SEEKER, i.e., seeker has spotted the runner and made it to the safe-area before
     * the runner.
     *
     * @param event
     * @param me whether this event speaks about your bot
     */
    @Override
    protected void hsRunnerCaptured(HSRunnerCaptured event, boolean me) {
        // runner - seeker je pikole, takze se radeji jeste schovame
        if (hide.isMeRunner() && hide.isMeAlive() && !me) {
            runnerRunToHide(4 * 4);
        }
        // runner - byl jsem zapikan
        else if (hide.isMeRunner() && hide.isMeAlive() && me) {
            --runnerMeGood;
        }

        // seeker - uz na runnera nemusime koukat
        if (hide.isMeSeeker()) {
            if (playerToCapture == this.players.getPlayer(event.getBotId())) {
                playerToCapture = null;
            }

            // koukneme jestli nekdo neni na druhe strane
            move.turnHorizontal(90);
        }
    }

    /**
     * Round state update has been received.
     * @param event
     */
    @Override
    protected void hsRoundState(HSRoundState event) {
    }

    /**
     * New round has just started, you may use this event to initialize your data-structures.
     * @param event
     */
    @Override
    protected void hsRoundStart(HSRoundStart event) {
        navigation.stopNavigation(); // prevent commands from previous round to be still effective

        starting = true;
        runnerHidden = false;
        runningToSafeArea = false;
        locationToHide = null;
        timeOutBeforeRunningToSafeArea = 4 * 4;

        playerToCapture = null;

        if (tabooPoints != null) {
            tabooPoints.clear();
        }
        
        // runneri nam utekli => zustaneme na pikole
        if(runnerSafe > 1) {
            seekerEndingStrategy = SeekerStrategy.StayAtSafeArea;
        }
        // runneri se schovavaji => budeme vice prohledavat
        else if(runnerSurvived > 1) {
            seekerEndingStrategy = SeekerStrategy.ExamineCoverPoints;
        }
        // runneri nam utikaji => zustaneme na pikole
        else if(runnerSafe >= 1 && runnerSurvived < 1) {
            seekerEndingStrategy = SeekerStrategy.StayAtSafeArea;
        }
        // zkusime radeji chodit pouze blizko pikoly
        else if(runnerSurvived == 1 || runnerSafe == 1) {
            seekerEndingStrategy = SeekerStrategy.NearSafeArea;
        }
        runnerSurvived = 0;
        runnerSafe = 0;
        
        roundTime = hide.getRemainingRoundTime();
    }

    /**
     * Round has just ended, you may use this event to cleanup your data-structures.
     * @param event
     */
    @Override
    protected void hsRoundEnd(HSRoundEnd event) {
    }

    /**
     * Some player (runner/seeker) score has just changed, you may examine the reason within {@link HSPlayerScoreChanged#getScoreChangeReason()}, see
     * {@link HSScoreChangeReason}.
     *
     * @param event
     * @param me whether this event speaks about your bot
     */
    @Override
    protected void hsPlayerScoreChanged(HSPlayerScoreChanged event, boolean me) {
    }

    /**
     * Some bot {@link HSBotState} has been updated.
     * @param event
     * @param me whether this event speaks about your bot
     */
    @Override
    protected void hsBotStateChanged(HSBotStateChanged event, boolean me) {
    }

    /**
     * SEEKER role has been just assigned to some bot.
     * @param event
     * @param me whether this event speaks about your bot
     */
    @Override
    protected void hsAssignSeeker(HSAssignSeeker event, boolean me) {
    }

    // -----
    // =====
    // LOGIC
    // =====
    // -----
    
    @Override
    public void beforeFirstLogic() {
        tabooPoints = new TabooSet<NavPoint>(bot);

        this.navigation.addStrongNavigationListener(new FlagListener<NavigationState>() {

            @Override
            public void flagChanged(NavigationState changedValue) {
                // bot se nekde zaseknul
                if (changedValue == NavigationState.STUCK) {
                    switch (hide.getMyState()) {
                        case RUNNER:
                            runnerStucked();
                            break;

                        case RUNNER_SPOTTED:
                            runnerStucked();
                            break;

                        case SEEKER:
                            seekerStucked();
                            break;
                    }
                }
                // seeker dobehnul na pozadovanou pozici
                else if(changedValue == NavigationState.TARGET_REACHED && hide.isMeSeeker()) {
                    // pokud to neni pikola tak ted urcite na pikolu chceme bezet
                    if(!hide.isInSafeArea(info.getLocation())) {
                        navigation.navigate(hide.getSafeArea());
                    }
                }
            }
        });
    }

    private void runnerStucked() {
        if (!runningToSafeArea) {
            // nekdo nam pravdepodobne prekazi v ceste, ale nikam nespechame, 
            // takze se otocime a pujdeme jinam
            navigation.stopNavigation();
            move.turnHorizontal(90);
            locationToHide = null;
        } else {
            // nekdo nam pravdepodobne prekazi v ceste, zkusime jednoduche vyhnuti
            move.dodgeLeft(info.getLocation().add(info.getVelocity()), false);
        }
    }

    private void seekerStucked() {
        if (hide.canCaptureRunner()) {
            // zkusime se prekazce vyhnout
            move.dodgeLeft(info.getLocation().add(info.getVelocity()), false);
        } else {
            // radeji se vratime do pikoly
            navigation.navigate(hide.getSafeArea());
        }
    }

    @Override
    public void logic() {
        // mark that another logic iteration has began
        log.info("====== LOGIC ITERATION ======");

        // UNCOMMENT TO GET DETAIL INFO DURING RUN-TIME
        // logState();

        if (!hide.isRoundRunning()) {
            // Game or Game-Round is not running yet...
            return;
        }

        if (!hide.isMeAlive()) {
            // I'm not alive ~ already captured / escaped
            return;
        }

        switch (hide.getMyState()) {
            case RUNNER:
                logicRunner();
                break;
            case RUNNER_SPOTTED:
                logicRunnerSpotted();
                break;
            case SEEKER:
                logicSeeker();
                break;
            default:
                // SHOULD NOT REACH HERE, BUT LET'S PLAY SAFE...
                return;
        }
    }

    private void logState() {
        log.info("------      STATE      ------");
        log.info("Game state:                  " + hide.getGameState());
        log.info("Game running:                " + hide.isGameRunning());
        if (!hide.isGameRunning()) {
            return;
        }
        log.info("Round running:               " + hide.isRoundRunning());
        if (!hide.isRoundRunning()) {
            return;
        }
        log.info("Me:                          " + hide.getMyState());
        log.info("Me spawned?                  " + hide.isMeAlive());
        log.info("Hiding time left:            " + hide.getRemainingHidingTime());
        log.info("Restricted area time left:   " + hide.getRemainingRestrictedAreaTime());
        log.info("Round time left:             " + hide.getRemainingRoundTime());
        log.info("Distance to safe-area:       " + hide.getMySafeAreaDistance());
        log.info("Distance to restricted-area: " + hide.getMyRestrictedAreaDistance());
    }

    private NavPoint getRandomAlmostFarthest(Collection<NavPoint> locations, NavPoint target, float tolerateDistance) {
        NavPoint farthestItem = null;
        float maxDistance = Float.MIN_VALUE;

        for (NavPoint item : locations) {
            float distance = fwMap.getDistance(item, target);
            if(distance == Float.POSITIVE_INFINITY) continue;
            if ((distance >= maxDistance && random.nextDouble() >= 0.8)
                    || ((distance + tolerateDistance) >= maxDistance && random.nextDouble() >= 0.5)) {
                maxDistance = distance;
                farthestItem = item;
            }
        }

        if (farthestItem == null && !locations.isEmpty()) {
            farthestItem = MyCollections.getRandom(locations);
        }

        return farthestItem;
    }
    
    private VisibilityLocation getRandomAlmostNearest(Collection<VisibilityLocation> locations, VisibilityLocation target, float tolerateDistance) {
        VisibilityLocation nearestItem = null;
        float minDistance = Float.MAX_VALUE;

        for (VisibilityLocation item : locations) {
            float distance = fwMap.getDistance(item.navPoint, target.navPoint);
            if(distance == Float.POSITIVE_INFINITY) continue;
            if ((distance <= minDistance && random.nextDouble() >= 0.8)
                    || ((distance - tolerateDistance) <= minDistance && random.nextDouble() >= 0.5)) {
                minDistance = distance;
                nearestItem = item;
            }
        }

        if (nearestItem == null && !locations.isEmpty()) {
            nearestItem = DistanceUtils.getNearest2D(locations, target);
        }

        return nearestItem;
    }

    private void setLocationHide(final boolean moveOnlySafe) {
        // najdeme mista, kam bychom mohli bezet
        Collection<NavPoint> pointsToRun;
        if (moveOnlySafe) {
            pointsToRun = tabooPoints.filter(visibility.getVisibleNavPoints());
        } else {
            pointsToRun = tabooPoints.filter(visibility.getCoverNavPointsFrom(hide.getSafeArea()));
        }

        // nechceme jit tam, kde uz jsme
        pointsToRun.remove(info.getNearestNavPoint());

        if (moveOnlySafe) {
            // nechceme jit na body, na ktere lze videt z pikoly
            pointsToRun = MyCollections.getFiltered(pointsToRun, new IFilter<NavPoint>() {

                @Override
                public boolean isAccepted(NavPoint object) {
                    return !visibility.isVisible(object, hide.getSafeArea());
                }
            });
        }

        // vezmeme nejaky vzdaleny bod
        if (!pointsToRun.isEmpty()) {
            locationToHide = getRandomAlmostFarthest(pointsToRun, info.getNearestNavPoint(), (pointsToRun.size() > 5) ? 1000 : 50);
        }
    }

    private void moveWhileHidden(final boolean moveOnlySafe) {
        if (locationToHide != null && navigation.isNavigating()) {
            return;
        }

        if (!navigation.isNavigating() || locationToHide == null) {
            setLocationHide(moveOnlySafe);
        }

        if (locationToHide != null) {
            // vyzkousime maximalne 10 jinych mist, kam jit
            for (int i = 0; i < 10; ++i) {
                if (locationToHide == null) {
                    break;
                }

                AStarResult<NavPoint> pathResult = this.aStar.findPath(info.getNearestNavPoint(), locationToHide, new IPFMapView<NavPoint>() {

                    @Override
                    public Collection<NavPoint> getExtraNeighbors(NavPoint node, Collection<NavPoint> mapNeighbors) {
                        return null;
                    }

                    @Override
                    public int getNodeExtraCost(NavPoint node, int mapCost) {
                        double extraCost = 0;

                        // radeji se pokusime vyhnout ostatnim hracum
                        if (!players.getVisiblePlayers().isEmpty()) {
                            for (Player player : players.getVisiblePlayers().values()) {
                                extraCost += 1000.0 / Location.getDistance2D(node.getLocation(), player.getLocation());
                            }
                        }

                        if (!moveOnlySafe) {
                            if (visibility.isVisible(node.getLocation(), hide.getSafeArea())) {
                                extraCost += 100;
                            }
                        }

                        return (int) extraCost;
                    }

                    @Override
                    public int getArcExtraCost(NavPoint nodeFrom, NavPoint nodeTo, int mapCost) {
                        return 0;
                    }

                    @Override
                    public boolean isNodeOpened(NavPoint node) {
                        if (moveOnlySafe) {
                            // nechceme byt videni z pikoly
                            return !visibility.isVisible(node.getLocation(), hide.getSafeArea());
                        } else {
                            return true;
                        }
                    }

                    @Override
                    public boolean isArcOpened(NavPoint nodeFrom, NavPoint nodeTo) {
                        if (moveOnlySafe) {
                            return !visibility.isVisible(nodeFrom.getLocation(), hide.getSafeArea()) && !visibility.isVisible(nodeTo.getLocation(), hide.getSafeArea());
                        } else {
                            return true;
                        }
                    }
                });

                if (pathResult.isSuccess()) {
                    // jdeme po nalezene ceste
                    navigation.navigate(new PrecomputedPathFuture<ILocated>(
                            pathResult.startNode,
                            pathResult.goalNode,
                            new ArrayList<ILocated>(pathResult.getPath())));

                    return;
                } else {
                    locationToHide = null;
                    tabooPoints.add(locationToHide, 5);
                    setLocationHide(moveOnlySafe);
                }
            }

            // nepodarilo se nam najit zadne misto, kam bezet, tak se trochu otocime
            this.move.turnHorizontal(45);
        } else {
            // nevidime zadne misto, kam bezet, tak se trochu otocime
            this.move.turnHorizontal(45);
        }
    }

    private void runnerRunToHide(final int timeout) {
        // pokud nas seeker nevidel, tak se pobezime schovat
        if (hide.getMyState() == HSBotState.RUNNER) {

            if (visibility.isVisible(hide.getSafeArea())) {
                // uz jsme blizko, tak tam doskocime
                if (hide.getSafeAreaDistance(info.getLocation()) < 400) {
                    move.dodgeTo(hide.getSafeArea(), true);
                    return;
                } 
                // seeker neni nablizku, tak se nemusime schovat
                else if (hide.getSeeker() != null && !hide.getSeeker().isVisible()) {
                    return;
                }
            }

            timeOutBeforeRunningToSafeArea += timeout;
            locationToHide = null;
            runningToSafeArea = false;

            if(runnerMeGood >= 0) {
                moveWhileHidden(false);
            } else {
                moveWhileHidden(true);
            }
        }
    }
    
    private void rotateRunner() {
        if (navigation.getPathExecutor() != null && navigation.isNavigating()) {
            runnerRotationVector = runnerRotationVector.rotateXY(Math.toRadians(45));
            navigation.getPathExecutor().setFocus(info.getLocation().add(runnerRotationVector));
        } else {
            move.turnHorizontal(65);
        }
    }
    
    // ------------
    // ============
    // LOGIC RUNNER
    // ============
    // ------------
    
    private boolean runningToSafeArea = false;
    private boolean runnerHidden = false;
    private NavPoint locationToHide = null;
    private int timeOutBeforeRunningToSafeArea = 0;
    private TabooSet<NavPoint> tabooPoints;
    private Location runnerRotationVector = new Location(400, 0);

    private void logicRunner() {
        log.info("------ RUNNER ------");
        // YOU ARE RUNNER => one who should be hiding

        if (hide.isHidingTime()) {
            log.info("HIDING TIME");
            // SEEKER has not been spawned yet, you have time to hide out!  

            if (!this.navigation.isNavigating()) {
                if (!runnerHidden) {
                    // bezime se nekam schovat
                    Collection<VisibilityLocation> coverPoints = this.visibility.getCoverPointsFrom(hide.getSafeArea());
                    navigation.navigate(MyCollections.getRandom(coverPoints));
                    runnerHidden = true;
                } else {
                    // pohybujeme se "schovani"
                    moveWhileHidden(true);
                }
            }
            rotateRunner();

            return;
        }

        if (hide.isRestrictedAreaActivated()) {
            log.info("RESTRICTED AREA ACTIVE");
            // SEEKER has been spawned into the environment but you cannot save yourself yet
            //         "easy-win" by standing in the safe are at the beginning of the game is not allowed 

            // pohybujeme se "schovani"
            moveWhileHidden(true);
            rotateRunner();

            return;
        }

        log.info("SAFE AREA ACTIVE");
        // SEEKER is present in the game, but restricted area is not active, 
        // you can save yourself by running to safe area before the SEEKER spots you and returns first

        if (!runningToSafeArea) {
            // vidime nejake hrace
            for (Player player : players.getVisiblePlayers().values()) {
                // vidime seekera, tak radeji rovnou pobezime na pikolu
                if (hide.getBotState(player.getId()) == HSBotState.SEEKER) {
                    timeOutBeforeRunningToSafeArea = 0;
                    runningToSafeArea = true;
                    navigation.navigate(this.hide.getSafeArea());
                    break;
                }
            }
        }

        if (timeOutBeforeRunningToSafeArea > 0) {
            // timeout pred behem na pikolu
            --timeOutBeforeRunningToSafeArea;
            moveWhileHidden(true);
            rotateRunner();
        } else if (!runningToSafeArea || !navigation.isNavigating()) {
            if(runnerMeGood >= 0 || (random.nextDouble() > 0.98 && runnerMeGood >= -5)) {
                // zkusime stesti a pobezime na pikolu
                runningToSafeArea = true;
                navigation.navigate(this.hide.getSafeArea());
            } else {
                rotateRunner();
            }
        }

        // jestli uz jsme blizko pikoly, tak si malinko pomuzeme skokem
        if (hide.getSafeAreaDistance(info.getLocation()) < 400) {
            move.dodgeTo(hide.getSafeArea(), true);
        }
    }

    // --------------------
    // ====================
    // LOGIC RUNNER SPOTTED
    // ====================
    // --------------------
    
    private void logicRunnerSpotted() {
        log.info("------ RUNNER SPOTTED ------");
        // YOU ARE RUNNER who has already been spotted by the SEEKER
        // if SEEKER gets to the safe point first, it will catch you

        // pikola jeste neni aktivni a seeker nas uz videl
        if (hide.isRestrictedAreaActivated()) {
            if (!runningToSafeArea || !navigation.isNavigating()) {
                // pobezime k pikole, pokud neni moc blizko
                if(fwMap.getDistance(info.getNearestNavPoint(), visibility.getNearestNavPointTo(hide.getSafeArea())) > hide.getGameConfig().getRestrictedAreaRadius() + 200) {
                    navigation.navigate(hide.getSafeArea());
                }
                runningToSafeArea = true;
            } 
            // uz jsme moc blizko pikoly, tak se radeji zastavime
            else if(navigation.isNavigating() && navigation.getRemainingDistance() < hide.getGameConfig().getRestrictedAreaRadius() + 100) {
                navigation.stopNavigation();
            }
            
            return;
        }
        
        if (!runningToSafeArea || !navigation.isNavigating()) {
            navigation.navigate(this.hide.getSafeArea());
            runningToSafeArea = true;
        }

        // jestli uz jsme blizko pikoly, tak si malinko pomuzeme skokem
        if (hide.getSafeAreaDistance(info.getLocation()) < 400) {
            move.dodgeTo(hide.getSafeArea(), true);
        }
    }
    
    // ------------
    // ============
    // LOGIC SEEKER
    // ============
    // ------------
    
    private enum SeekerStrategy { StayAtSafeArea, NearSafeArea, ExamineCoverPoints }
    
    private boolean starting = true;
    private Location seekerRotationVector = new Location(400, 0);
    private Player playerToCapture = null;
    private SeekerStrategy seekerEndingStrategy = SeekerStrategy.StayAtSafeArea;

    private void logicSeeker() {
        log.info("------ SEEKER ------");
        // you are SEEKER, one who should be finding other runners and catching them

        if (hide.canCaptureRunner()) {
            log.info("CAN CAPTURE RUNNER!");
            // you have had spotted some runner that have not made it to the safe area yet

            if (!this.navigation.isNavigating() && !hide.isInSafeArea(info.getLocation())) {
                this.navigation.navigate(this.hide.getSafeArea());
            }

            // koukame na runnera
            if (playerToCapture != null) {
                if (hide.getBotState(playerToCapture.getId()) == HSBotState.RUNNER_SPOTTED) {
                    playerToCapture = null;
                } else if(navigation.getPathExecutor() != null && navigation.isNavigating()) {
                    navigation.getPathExecutor().setFocus(playerToCapture.getLocation());
                } else {
                    move.turnTo(playerToCapture);
                }
            }

            // otacime se, treba objevime dalsiho runnera
            if (playerToCapture == null && navigation.getPathExecutor() != null && navigation.isNavigating()) {
                seekerRotationVector = seekerRotationVector.rotateXY(Math.toRadians(65));
                navigation.getPathExecutor().setFocus(info.getLocation().add(seekerRotationVector));
            } else if (playerToCapture == null) {
                move.turnHorizontal(65);
            }

            // vidime nejake hrace
            if (playerToCapture == null || hide.getBotState(playerToCapture.getId()) == HSBotState.RUNNER_SPOTTED) {
                for (Player player : players.getVisiblePlayers().values()) {
                    // vidime noveho runnera
                    if (hide.getBotState(player.getId()) == HSBotState.RUNNER) {
                        playerToCapture = player;
                        if(navigation.getPathExecutor() != null && navigation.isNavigating()) {
                            navigation.getPathExecutor().setFocus(playerToCapture.getLocation());
                        } else {
                            move.turnTo(playerToCapture);
                        }
                        break;
                    }
                }
            }

            // jestli uz jsme blizko pikoly, tak si malinko pomuzeme skokem
            if (!hide.isInSafeArea(info.getLocation()) && hide.getSafeAreaDistance(info.getLocation()) < 400) {
                move.dodgeTo(hide.getSafeArea(), true);
            }
        } else {
            // you cannot capture any runner by returning to the safe area,
            // you should try to spot some first    

            if (!this.navigation.isNavigating()) {
                Collection<VisibilityLocation> pointsToGo;
                
                // na zacatku pobezime prozkoumat nektery blizky cover point
                if(starting) {
                    pointsToGo = this.visibility.getCoverPointsFrom(hide.getSafeArea());
                    navigation.navigate(getRandomAlmostNearest(pointsToGo, visibility.getNearestVisibilityLocation(), 100));
                    starting = false;
                }
                // mame hodne casu do konce kola, takze budeme zly a budeme stat na pikole
                else if(hide.getRemainingHidingTime() / roundTime >= 0.3) {
                    if(!hide.isInSafeArea(info.getLocation())) {
                        navigation.navigate(hide.getSafeArea());
                    }
                } 
                // blizi se konec
                else {
                    // budeme stat na pikole
                    if(seekerEndingStrategy == SeekerStrategy.StayAtSafeArea) {
                        if(!hide.isInSafeArea(info.getLocation())) {
                            navigation.navigate(hide.getSafeArea());
                        }
                    }
                    // budeme stale blizko pikoly (pobezime na mista, z kterych uvidime na pikolu)
                    else if(seekerEndingStrategy == SeekerStrategy.NearSafeArea) {
                        pointsToGo = this.visibility.getVisiblePointsFrom(hide.getSafeArea());
                        pointsToGo.remove(this.visibility.getNearestVisibilityLocation());
                        navigation.navigate(getRandomAlmostNearest(pointsToGo, visibility.getNearestVisibilityLocation(), 1000));
                    }
                    // pobezime prozkoumat nektery cover point
                    else if(seekerEndingStrategy == SeekerStrategy.ExamineCoverPoints) {
                        pointsToGo = this.visibility.getCoverPointsFrom(hide.getSafeArea());
                        navigation.navigate(MyCollections.getRandom(pointsToGo));
                    }
                }
            }

            // otacime se, treba objevime runnera
            if (navigation.getPathExecutor() != null && navigation.isNavigating()) {
                seekerRotationVector = seekerRotationVector.rotateXY(Math.toRadians(45));
                navigation.getPathExecutor().setFocus(info.getLocation().add(seekerRotationVector));
            } else {
                move.turnHorizontal(65);
            }

            // vidime nejake hrace
            for (Player player : players.getVisiblePlayers().values()) {
                // objevili jsme runnera
                if (hide.getBotState(player.getId()) == HSBotState.RUNNER) {
                    playerToCapture = player;
                    seekerRunToSafeArea();
                    break;
                }
            }
        }
    }

    private void seekerRunToSafeArea() {
        if (navigation.getPathExecutor() != null) {
            this.navigation.getPathExecutor().setFocus(null);
        }
        
        if(!hide.isInSafeArea(info.getLocation())) {
            this.navigation.navigate(this.hide.getSafeArea());
        }

        if (playerToCapture != null) {
            if(navigation.getPathExecutor() != null && navigation.isNavigating()) {
                navigation.getPathExecutor().setFocus(playerToCapture.getLocation());
            } else {
                move.turnTo(playerToCapture);
            }
        }
    }

    @Override
    public void botKilled(BotKilled event) {
        navigation.stopNavigation();
    }

    public static void main(String args[]) throws PogamutException {
        // LEAVE THIS BE!
        // Start the debug game via HideAndSeekGame !!!
        new UT2004BotRunner(HideAndSeekBot.class, "HSBot").setMain(true).setLogLevel(Level.WARNING).startAgents(1);
    }
}
